use serde::{Deserialize, Serialize};
use smallvec::SmallVec;
use specs::{
    error::NoError,
    saveload::{ConvertSaveload, Marker},
    Component, Entity, HashMapStorage,
};

use crate::misc::FlaggedStorage;

#[derive(Clone, Debug, Default)]
pub struct Orbit {
    pub min: f32,
    pub max: f32,
    pub fleets: SmallVec<[Option<Entity>; 8]>,
}

#[derive(Serialize, Deserialize)]
pub struct OrbitData<M> {
    pub min: f32,
    pub max: f32,
    pub fleets: Vec<Option<M>>,
}

#[derive(Copy, Clone, Debug)]
pub struct Owned {
    pub owner: Entity,
}

#[derive(Serialize, Deserialize)]
pub struct OwnedData<M> {
    pub owner: M,
}

impl Component for Orbit {
    type Storage = HashMapStorage<Self>;
}

impl<M> ConvertSaveload<M> for Orbit
where
    for<'de> M: Marker + Serialize + Deserialize<'de>,
{
    type Data = OrbitData<M>;
    type Error = NoError;

    fn convert_into<F>(&self, mut ids: F) -> Result<Self::Data, Self::Error>
    where
        F: FnMut(Entity) -> Option<M>,
    {
        let Orbit { min, max, fleets } = self;

        let min = *min;
        let max = *max;
        let fleets = fleets
            .iter()
            .cloned()
            .map(|e| e.and_then(|e| ids(e)))
            .collect();

        Ok(OrbitData { min, max, fleets })
    }

    fn convert_from<F>(data: Self::Data, mut ids: F) -> Result<Self, Self::Error>
    where
        F: FnMut(M) -> Option<Entity>,
    {
        let OrbitData { min, max, fleets } = data;
        let fleets = fleets.into_iter().map(|e| e.and_then(|e| ids(e))).collect();

        Ok(Orbit { min, max, fleets })
    }
}

impl Component for Owned {
    type Storage = FlaggedStorage<Self, HashMapStorage<Self>>;
}

impl<M> ConvertSaveload<M> for Owned
where
    for<'de> M: Marker + Serialize + Deserialize<'de>,
{
    type Data = OwnedData<M>;
    type Error = NoError;

    fn convert_into<F>(&self, mut ids: F) -> Result<Self::Data, Self::Error>
    where
        F: FnMut(Entity) -> Option<M>,
    {
        let owner = ids(self.owner).unwrap();

        Ok(OwnedData { owner })
    }

    fn convert_from<F>(data: Self::Data, mut ids: F) -> Result<Self, Self::Error>
    where
        F: FnMut(M) -> Option<Entity>,
    {
        let owner = ids(data.owner).unwrap();

        Ok(Owned { owner })
    }
}
